@@ -70,6 +70,24 @@
 // FIXME (v2): do include dir properly.
 #include "vendor/mcmc/mcmc.h"
 
+enum mcp_memprofile_types {
+    mcp_memp_free = 0,
+    mcp_memp_string,
+    mcp_memp_table,
+    mcp_memp_func,
+    mcp_memp_userdata,
+    mcp_memp_thread,
+    mcp_memp_default,
+    mcp_memp_realloc,
+};
+
+struct mcp_memprofile {
+    struct timespec last_status; // for per-second prints on status
+    int id;
+    uint64_t allocs[8];
+    uint64_t alloc_bytes[8];
+};
+
 // Note: value created from thin air. Could be shorter.
 #define MCP_REQUEST_MAXLEN KEY_MAX_LENGTH * 2
 
@@ -80,7 +98,12 @@
 
 #define MCP_YIELD_POOL 1
 #define MCP_YIELD_AWAIT 2
-#define MCP_YIELD_LOCAL 3
+#define MCP_YIELD_INTERNAL 3
+#define MCP_YIELD_WAITCOND 4
+#define MCP_YIELD_WAITHANDLE 5
+
+#define SHAREDVM_FGEN_IDX 1
+#define SHAREDVM_FGENSLOT_IDX 2
 
 // all possible commands.
 #define CMD_FIELDS \
@@ -181,18 +204,20 @@ struct proxy_tunables {
     struct timeval connect;
     struct timeval retry; // wait time before retrying a dead backend
     struct timeval read;
-#ifdef HAVE_LIBURING
-    struct __kernel_timespec connect_ur;
-    struct __kernel_timespec retry_ur;
-    struct __kernel_timespec read_ur;
-#endif // HAVE_LIBURING
+    struct timeval flap; // need to stay connected this long or it's flapping
+    float flap_backoff_ramp; // factorial for retry time
+    uint32_t flap_backoff_max; // don't backoff longer than this.
     int backend_failure_limit;
+    int max_ustats; // limit the ustats index.
     bool tcp_keepalive;
+    bool use_iothread; // default for using the bg io thread.
+    bool down; // backend is forced into a down/bad state.
 };
 
 typedef STAILQ_HEAD(pool_head_s, mcp_pool_s) pool_head_t;
 typedef struct {
-    lua_State *proxy_state;
+    lua_State *proxy_state; // main configuration vm
+    lua_State *proxy_sharedvm; // sub VM for short-lock global events/data
     void *proxy_code;
     proxy_event_thread_t *proxy_io_thread;
     uint64_t active_req_limit; // max total in-flight requests
@@ -205,11 +230,14 @@ typedef struct {
     pthread_t manager_tid; // deallocation management thread
     pthread_mutex_t manager_lock;
     pthread_cond_t manager_cond;
+    pthread_mutex_t sharedvm_lock; // protect statevm above
     pool_head_t manager_head; // stack for pool deallocation.
     bool worker_done; // signal variable for the worker lock/cond system.
     bool worker_failed; // covered by worker_lock as well.
     bool use_uring; // use IO_URING for backend connections.
     bool loading; // bool indicating an active config load.
+    bool memprofile; // indicate if we want to profile lua memory.
+    uint8_t memprofile_thread_counter;
     struct proxy_global_stats global_stats;
     struct proxy_user_stats user_stats;
     struct proxy_tunables tunables; // NOTE: updates covered by stats_lock
@@ -221,13 +249,18 @@ typedef struct {
 // Operations from the config VM don't have a libevent thread.
 #define PROXY_GET_CTX(L) (*(proxy_ctx_t **)lua_getextraspace(L))
 
+struct proxy_hook_ref {
+    int lua_ref;
+    void *ctx; // if we're a generator based function.
+};
+
 struct proxy_hook_tagged {
     uint64_t tag;
-    int lua_ref;
+    struct proxy_hook_ref ref;
 };
 
 struct proxy_hook {
-    int lua_ref;
+    struct proxy_hook_ref ref;
     int tagcount;
     struct proxy_hook_tagged *tagged; // array of possible tagged hooks.
 };
@@ -259,6 +292,8 @@ typedef struct mcp_backend_label_s mcp_backend_label_t;
 typedef struct mcp_backend_s mcp_backend_t;
 typedef struct mcp_request_s mcp_request_t;
 typedef struct mcp_parser_s mcp_parser_t;
+typedef struct mcp_rcontext_s mcp_rcontext_t;
+typedef struct mcp_funcgen_s mcp_funcgen_t;
 
 #define PARSER_MAX_TOKENS 24
 
@@ -277,6 +312,7 @@ struct mcp_parser_s {
     uint8_t keytoken; // because GAT. sigh. also cmds without a key.
     uint32_t parsed; // how far into the request we parsed already
     uint32_t reqlen; // full length of request buffer.
+    uint32_t endlen; // index to the start of \r\n or \n
     int vlen;
     uint32_t klen; // length of key.
     uint16_t tokens[PARSER_MAX_TOKENS]; // offsets for start of each token
@@ -292,7 +328,6 @@ struct mcp_parser_s {
 #define MAX_REQ_TOKENS 2
 struct mcp_request_s {
     mcp_parser_t pr; // non-lua-specific parser handling.
-    mcp_backend_t *be; // backend handling this request.
     bool ascii_multiget; // ascii multiget mode. (hide errors/END)
     char request[];
 };
@@ -315,6 +350,7 @@ struct mcp_backend_label_s {
     char port[MAX_PORTLEN+1];
     char label[MAX_LABELLEN+1];
     size_t llen; // cache label length for small speedup in pool creation.
+    int conncount; // number of sockets to make.
     struct proxy_tunables tunables;
 };
 
@@ -327,16 +363,15 @@ struct mcp_backend_wrap_s {
     mcp_backend_t *be;
 };
 
-// FIXME: inline the mcmc client data.
-// TODO: event_thread -> something? union of owner type?
-struct mcp_backend_s {
+struct mcp_backendconn_s {
+    mcp_backend_t *be_parent; // find the wrapper.
+    int self; // our index into the parent array.
     int depth; // total number of requests in queue
     int pending_read; // number of requests written to socket, pending read.
     int failed_count; // number of fails (timeouts) in a row
+    int flap_count; // number of times we've "flapped" into bad state.
     proxy_event_thread_t *event_thread; // event thread owning this backend.
     void *client; // mcmc client
-    STAILQ_ENTRY(mcp_backend_s) be_next; // stack for backends
-    STAILQ_ENTRY(mcp_backend_s) beconn_next; // stack for connecting conns
     io_head_t io_head; // stack of requests.
     io_pending_proxy_t *io_next; // next request to write.
     char *rbuf; // statically allocated read buffer.
@@ -345,23 +380,32 @@ struct mcp_backend_s {
     struct event write_event; // libevent: only used when socket wbuf full
     struct event timeout_event; // libevent: alarm for pending reads
     struct proxy_tunables tunables;
-#ifdef HAVE_LIBURING
-    proxy_event_t ur_rd_ev; // liburing.
-    proxy_event_t ur_wr_ev; // need a separate event/cb for writing/polling
-    proxy_event_t ur_te_ev; // for timeout handling
-#endif
+    struct timeval last_failed; // time the backend was last reset.
     enum mcp_backend_states state; // readback state machine
     int connect_flags; // flags to pass to mcmc_connect
-    bool transferred; // if beconn has been shipped to owner thread.
     bool connecting; // in the process of an asynch connection.
     bool validating; // in process of validating a new backend connection.
     bool can_write; // recently got a WANT_WRITE or are connecting.
-    bool stacked; // if backend already queued for syscalls.
     bool bad; // timed out, marked as bad.
-    bool use_io_thread; // note if this backend is worker-local or not.
     struct iovec write_iovs[BE_IOV_MAX]; // iovs to stage batched writes
+};
+
+// TODO: move depth and flags to a second top level array so we can make index
+// decisions from fewer memory stalls.
+struct mcp_backend_s {
+    int conncount; // total number of connections managed.
+    int depth; // temporary depth counter for io_head
+    bool transferred; // if beconn has been shipped to owner thread.
+    bool use_io_thread; // note if this backend is worker-local or not.
+    bool stacked; // if backend already queued for syscalls.
+    STAILQ_ENTRY(mcp_backend_s) beconn_next; // stack for connecting conns
+    STAILQ_ENTRY(mcp_backend_s) be_next; // stack for backends
+    io_head_t io_head; // stack of inbound requests.
     char name[MAX_NAMELEN+1];
     char port[MAX_PORTLEN+1];
+    char label[MAX_LABELLEN+1];
+    struct proxy_tunables tunables; // this gets copied a few times for speed.
+    struct mcp_backendconn_s be[];
 };
 typedef STAILQ_HEAD(be_head_s, mcp_backend_s) be_head_t;
 typedef STAILQ_HEAD(beconn_head_s, mcp_backend_s) beconn_head_t;
@@ -431,12 +475,13 @@ struct _io_pending_proxy_t {
     mc_resp *resp;
     io_queue_cb return_cb; // called on worker thread.
     io_queue_cb finalize_cb; // called back on the worker thread.
+    STAILQ_ENTRY(io_pending_t) iop_next; // queue chain.
     // original struct ends here
 
-    int io_type; // extstore IO or backend IO
-    int coro_ref; // lua registry reference to the coroutine
-    lua_State *coro; // pointer directly to the coroutine
+    mcp_rcontext_t *rctx; // pointer to request context.
+    int queue_handle; // queue slot to return this result to
     bool ascii_multiget; // passed on from mcp_r_t
+    uint8_t io_type; // extstore IO or backend IO
     union {
         // extstore IO.
         struct {
@@ -451,19 +496,21 @@ struct _io_pending_proxy_t {
         };
         // backend request IO
         struct {
+            // FIXME: use top level next chain
             struct _io_pending_proxy_t *next; // stack for IO submission
             STAILQ_ENTRY(_io_pending_proxy_t) io_next; // stack for backends
-            int mcpres_ref; // mcp.res reference used for await()
             mcp_backend_t *backend; // backend server to request from
             struct iovec iov[2]; // request string + tail buffer
             int iovcnt; // 1 or 2...
             unsigned int iovbytes; // total bytes in the iovec
+            int mcpres_ref; // mcp.res reference used for await()
             int await_ref; // lua reference if we were an await object
             mcp_resp_t *client_resp; // reference (currently pointing to a lua object)
             bool flushed; // whether we've fully written this request to a backend.
             bool is_await; // are we an await object?
             bool await_first; // are we the main route for an await object?
             bool await_background; // dummy IO for backgrounded awaits
+            bool qcount_incr; // HACK.
         };
     };
 };
@@ -492,6 +539,7 @@ struct mcp_pool_s {
     int phc_ref;
     int self_ref; // TODO (v2): double check that this is needed.
     int pool_size;
+    int pool_be_total; // can be different from pool size for worker IO
     bool use_iothread;
     mcp_pool_be_t pool[];
 };
@@ -503,11 +551,16 @@ typedef struct {
 
 // utils
 bool proxy_bufmem_checkadd(LIBEVENT_THREAD *t, int len);
+void mcp_sharedvm_delta(proxy_ctx_t *ctx, int tidx, const char *name, int delta);
 
 // networking interface
 void proxy_init_event_thread(proxy_event_thread_t *t, proxy_ctx_t *ctx, struct event_base *base);
 void *proxy_event_thread(void *arg);
 void proxy_run_backend_queue(be_head_t *head);
+struct mcp_backendconn_s *proxy_choose_beconn(mcp_backend_t *be);
+mcp_resp_t *mcp_prep_resobj(lua_State *L, mcp_request_t *rq, mcp_backend_t *be, LIBEVENT_THREAD *t);
+mcp_resp_t *mcp_prep_bare_resobj(lua_State *L, LIBEVENT_THREAD *t);
+io_pending_proxy_t *mcp_queue_rctx_io(mcp_rcontext_t *rctx, mcp_request_t *rq, mcp_backend_t *be, mcp_resp_t *r);
 
 // await interface
 enum mcp_await_e {
@@ -520,19 +573,162 @@ enum mcp_await_e {
 };
 int mcplib_await(lua_State *L);
 int mcplib_await_logerrors(lua_State *L);
-int mcplib_await_run(conn *c, mc_resp *resp, lua_State *L, int coro_ref);
+int mcplib_await_run_rctx(mcp_rcontext_t *rctx);
 int mcplib_await_return(io_pending_proxy_t *p);
 
 // internal request interface
 int mcplib_internal(lua_State *L);
-int mcplib_internal_run(lua_State *L, conn *c, mc_resp *top_resp, int coro_ref);
+int mcplib_internal_run(mcp_rcontext_t *rctx);
 
 // user stats interface
+#define MAX_USTATS_DEFAULT 1024
 int mcplib_add_stat(lua_State *L);
 int mcplib_stat(lua_State *L);
 size_t _process_request_next_key(mcp_parser_t *pr);
 int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen);
 mcp_request_t *mcp_new_request(lua_State *L, mcp_parser_t *pr, const char *command, size_t cmdlen);
+void mcp_set_request(mcp_parser_t *pr, mcp_request_t *r, const char *command, size_t cmdlen);
+
+// rate limit interfaces
+int mcplib_ratelim_tbf(lua_State *L);
+int mcplib_ratelim_tbf_call(lua_State *L);
+
+// request function generator interface
+void proxy_return_rctx_cb(io_pending_t *pending);
+void proxy_finalize_rctx_cb(io_pending_t *pending);
+
+enum mcp_rqueue_e {
+    QWAIT_IDLE = 0,
+    QWAIT_ANY,
+    QWAIT_OK,
+    QWAIT_GOOD,
+    QWAIT_FASTGOOD,
+    QWAIT_HANDLE
+};
+
+enum mcp_funcgen_router_e {
+    FGEN_ROUTER_NONE = 0,
+    FGEN_ROUTER_SHORTSEP,
+    FGEN_ROUTER_LONGSEP,
+    FGEN_ROUTER_ANCHORSM,
+    FGEN_ROUTER_ANCHORBIG,
+};
+
+struct mcp_router_long_s {
+    char start[KEY_HASH_FILTER_MAX+1];
+    char stop[KEY_HASH_FILTER_MAX+1];
+};
+
+struct mcp_funcgen_router {
+    enum mcp_funcgen_router_e type;
+    union {
+        char sep;
+        char lsep[KEY_HASH_FILTER_MAX+1];
+        char anchorsm[2]; // short anchored mode.
+        struct mcp_router_long_s big;
+    } conf;
+    mcp_funcgen_t *def_fgen; // default route
+    int map_ref;
+};
+
+struct mcp_funcgen_s {
+    LIBEVENT_THREAD *thread; // worker thread that created this funcgen.
+    int generator_ref; // reference to the generator function.
+    int self_ref; // self-reference if we're attached anywhere
+    int argument_ref; // reference to an argument to pass to generator
+    int name_ref; // reference to string name for the generator
+    int max_queues; // how many queue slots rctx's have
+    unsigned int refcount; // reference counter
+    unsigned int total; // total contexts managed
+    unsigned int free; // free contexts
+    unsigned int free_max; // size of list below.
+    unsigned int routecount; // total routes if this fgen is a router.
+    bool closed; // the hook holding this fgen has been replaced
+    bool ready; // if we're locked down or not.
+    mcp_rcontext_t **list;
+    struct mcp_rqueue_s *queue_list;
+    struct mcp_funcgen_router router;
+};
+
+#define RQUEUE_TYPE_NONE 0
+#define RQUEUE_TYPE_POOL 1
+#define RQUEUE_TYPE_FGEN 2
+#define RQUEUE_ASSIGNED (1<<0)
+#define RQUEUE_R_RESUME (1<<1)
+#define RQUEUE_R_GOOD (1<<3)
+#define RQUEUE_R_OK (1<<4)
+#define RQUEUE_R_ANY (1<<5)
+#define RQUEUE_R_ERROR (1<<7)
+
+enum mcp_rqueue_state {
+    RQUEUE_IDLE = 0,
+    RQUEUE_QUEUED,
+    RQUEUE_ACTIVE,
+    RQUEUE_COMPLETE,
+    RQUEUE_WAITED
+};
+
+struct mcp_rqueue_s {
+    int obj_ref; // reference to pool/func/etc object
+    int cb_ref; // if a lua callback was specified
+    int req_ref; // reference to associated request object.
+    int res_ref; // reference to lua response object.
+    void *obj; // direct pointer to the object for fast access.
+    mcp_request_t *rq; // request set to this slot
+    mcp_resp_t *res_obj; // pointer to result object
+    enum mcp_rqueue_state state; // queued/active/etc
+    uint8_t obj_type; // what the obj_ref actually is.
+    uint8_t flags; // bit flags for various states
+};
+
+struct mcp_rcontext_s {
+    int request_ref; // top level request for this context.
+    int function_ref; // ref to the created route function.
+    int coroutine_ref; // ref to our encompassing coroutine.
+    unsigned int async_pending; // legacy async handling
+    int pending_reqs; // pending requests and sub-requests
+    unsigned int wait_count;
+    unsigned int wait_done; // TODO: change these variables to uint8's
+    int wait_handle; // waiting on a specific queue slot
+    int parent_handle; // queue slot in parent rctx
+    enum mcp_rqueue_e wait_mode;
+    bool first_queue; // HACK
+    lua_State *Lc; // coroutine thread pointer.
+    mcp_request_t *request; // ptr to the above reference.
+    mcp_rcontext_t *parent; // parent rctx in the call graph
+    conn *c; // associated client object.
+    mc_resp *resp; // top level response object to fill.
+    mcp_funcgen_t *fgen; // parent function generator context.
+    struct mcp_rqueue_s qslots[]; // queueable slots.
+};
+
+void mcp_run_rcontext_handle(mcp_rcontext_t *rctx, int handle);
+void mcp_process_rctx_wait(mcp_rcontext_t *rctx, int handle);
+int mcp_process_rqueue_return(mcp_rcontext_t *rctx, int handle, mcp_resp_t *res);
+int mcplib_rcontext_handle_set_cb(lua_State *L);
+int mcplib_rcontext_enqueue(lua_State *L);
+int mcplib_rcontext_wait_cond(lua_State *L);
+int mcplib_rcontext_wait_handle(lua_State *L);
+int mcplib_rcontext_enqueue_and_wait(lua_State *L);
+int mcplib_rcontext_res_good(lua_State *L);
+int mcplib_rcontext_res_any(lua_State *L);
+int mcplib_rcontext_res_ok(lua_State *L);
+int mcplib_rcontext_result(lua_State *L);
+int mcplib_funcgenbare_new(lua_State *L);
+int mcplib_funcgen_new(lua_State *L);
+int mcplib_funcgen_new_handle(lua_State *L);
+int mcplib_funcgen_ready(lua_State *L);
+int mcplib_router_new(lua_State *L);
+mcp_rcontext_t *mcp_funcgen_start(lua_State *L, mcp_funcgen_t *fgen, mcp_parser_t *pr);
+mcp_rcontext_t *mcp_funcgen_get_rctx(lua_State *L, int fgen_ref, mcp_funcgen_t *fgen);
+void mcp_funcgen_return_rctx(mcp_rcontext_t *rctx);
+int mcplib_funcgen_gc(lua_State *L);
+void mcp_funcgen_reference(lua_State *L);
+void mcp_funcgen_dereference(lua_State *L, mcp_funcgen_t *fgen);
+void mcp_rcontext_push_rqu_res(lua_State *L, mcp_rcontext_t *rctx, int handle);
+
+
+int mcplib_factory_command_new(lua_State *L);
 
 // request interface
 int mcplib_request(lua_State *L);
@@ -545,16 +741,26 @@ int mcplib_request_ntokens(lua_State *L);
 int mcplib_request_has_flag(lua_State *L);
 int mcplib_request_flag_token(lua_State *L);
 int mcplib_request_gc(lua_State *L);
+void mcp_request_cleanup(LIBEVENT_THREAD *t, mcp_request_t *rq);
+
+// response interface
+void mcp_response_cleanup(LIBEVENT_THREAD *t, mcp_resp_t *r);
+void mcp_set_resobj(mcp_resp_t *r, mcp_request_t *rq, mcp_backend_t *be, LIBEVENT_THREAD *t);
 
 int mcplib_open_dist_jump_hash(lua_State *L);
 int mcplib_open_dist_ring_hash(lua_State *L);
 
-int proxy_run_coroutine(lua_State *Lc, mc_resp *resp, io_pending_proxy_t *p, conn *c);
-mcp_backend_t *mcplib_pool_proxy_call_helper(lua_State *L, mcp_pool_proxy_t *pp, const char *key, size_t len);
-void mcp_request_attach(lua_State *L, mcp_request_t *rq, io_pending_proxy_t *p);
+int proxy_run_rcontext(mcp_rcontext_t *rctx);
+mcp_backend_t *mcplib_pool_proxy_call_helper(mcp_pool_proxy_t *pp, const char *key, size_t len);
+void mcp_request_attach(mcp_request_t *rq, io_pending_proxy_t *p);
 int mcp_request_render(mcp_request_t *rq, int idx, const char *tok, size_t len);
 void proxy_lua_error(lua_State *L, const char *s);
-void proxy_lua_ferror(lua_State *L, const char *fmt, ...);
+#define proxy_lua_ferror(L, fmt, ...) \
+    do { \
+        lua_pushfstring(L, fmt, __VA_ARGS__); \
+        lua_error(L); \
+    } while (0)
+
 #define PROXY_SERVER_ERROR "SERVER_ERROR "
 #define PROXY_CLIENT_ERROR "CLIENT_ERROR "
 void proxy_out_errstring(mc_resp *resp, char *type, const char *str);
@@ -564,5 +770,5 @@ int proxy_thread_loadconf(proxy_ctx_t *ctx, LIBEVENT_THREAD *thr);
 // TODO (v2): more .h files, perhaps?
 int mcplib_open_hash_xxhash(lua_State *L);
 
-__attribute__((unused)) void dump_stack(lua_State *L);
+__attribute__((unused)) void dump_stack(lua_State *L, const char *msg);
 #endif
